"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
from langmuir import *
from scipy.constants import value as constants
from pytest import approx
import pytest
import numpy as np

def test_species_defaults():
    T = 1000
    n = 1e11
    e = constants('elementary charge')
    m = constants('electron mass')
    k = constants('Boltzmann constant')
    s = Species(n=n, T=T)
    assert(s.m == approx(m))
    assert(s.q == approx(-e))
    assert(s.n == approx(n))
    assert(s.T == approx(T))
    assert(s.vth == approx(np.sqrt(k*1000/m)))
    assert(s.alpha == 0)
    assert(s.kappa == float('inf'))

def test_species_electron():
    s = Species('electron', n=1e11, T=1000)
    assert(s.m == approx(constants('electron mass')))
    assert(s.q == approx(-constants('elementary charge')))

def test_species_proton():
    s = Species('proton', n=1e11, T=1000)
    assert(s.m == approx(constants('proton mass')))
    assert(s.q == approx(constants('elementary charge')))

def test_species_positron():
    s = Species('positron', n=1e11, T=1000)
    assert(s.m == approx(constants('electron mass')))
    assert(s.q == approx(constants('elementary charge')))

def test_species_maxwellian():
    s = Species(n=1e11, T=1000)
    assert(s.kappa == float('inf'))
    assert(s.alpha == approx(0))

def test_species_kappa():
    s = Species(n=1e11, T=1000, kappa=4)
    assert(s.kappa == approx(4))
    assert(s.alpha == approx(0))

def test_species_cairns():
    s = Species(n=1e11, T=1000, alpha=0.2)
    assert(s.kappa == float('inf'))
    assert(s.alpha == approx(0.2))

def test_species_kappa_cairns():
    s = Species(n=1e11, T=1000, kappa=4, alpha=0.2)
    assert(s.kappa == approx(4))
    assert(s.alpha == approx(0.2))

def test_species_q():
    s = Species(n=1e11, T=1000, q=123)
    assert(s.q == approx(123))

def test_species_Z():
    s = Species(n=1e11, T=1000, Z=2)
    assert(s.q == approx(2*constants('elementary charge')))

def test_species_m():
    s = Species(n=1e11, T=1000, m=234)
    assert(s.m == approx(234))

def test_species_amu():
    s = Species(n=1e11, T=1000, amu=16)
    assert(s.m == approx(16*constants('atomic mass constant')))

def test_species_vth():
    s = Species(n=1e11, vth=1e6)
    assert(s.T == approx(1e6**2*s.m/constants('Boltzmann constant')))
    assert(s.vth == approx(1e6))

def test_species_eV():
    s = Species(n=1e11, eV=0.2)
    e = constants('elementary charge')
    k = constants('Boltzmann constant')
    assert(s.T == approx(0.2*e/k))

def test_species_repr():
    # A good representation of a (small) object is one you can use to recreate
    # the object
    s  = Species('kappa-cairns', q=1, m=3, n=1000, T=1000, kappa=6, alpha=0.2)
    s2 = eval(str(s))
    assert(s2.q == approx(s.q))
    assert(s2.m == approx(s.m))
    assert(s2.n == approx(s.n))
    assert(s2.T == approx(s.T))
    assert(s2.kappa == approx(s.kappa))
    assert(s2.alpha == approx(s.alpha))

def test_species_convenience_values():
    s = Species(n=1e11, T=1000)
    omega_p = 17839863.646512896
    eV = 0.08617330337217212
    vth = 123111.06021431087
    assert(s.omega_p == approx(omega_p))
    assert(s.freq_p == approx(omega_p/(2*np.pi)))
    assert(s.period_p == approx(2*np.pi/omega_p))
    assert(s.eV == approx(eV))
    assert(s.vth == approx(vth))

def test_species_convenience_methods():
    s = Species(n=1e11, T=1000)
    B = 50e-6
    omega_c = 8794100.11801062
    assert(s.omega_c(B) == approx(omega_c))
    assert(s.freq_c(B) == approx(omega_c/(2*np.pi)))
    assert(s.period_c(B) == approx(2*np.pi/omega_c))
    assert(s.larmor(B) == approx(s.vth/omega_c))

def test_species_debye_length():
    sp_m = Species(n=1e11, T=1000)
    sp_k = Species('kappa', n=1e11, T=1000, kappa=4)
    sp_a = Species('cairns', n=1e11, T=1000, alpha=0.2)
    sp_ka = Species('kappa-cairns', n=1e11, T=1000, kappa=4, alpha=0.2)
    assert(sp_m.debye == approx(6.900896931e-03))
    assert(sp_k.debye == approx(0.0058323223992018209))
    assert(sp_a.debye == approx(0.010911276093510305))
    assert(sp_ka.debye == approx(0.011952704095164884))

def test_debye():
    plasma = []
    plasma.append(Species('electron', n=1e11, T=1000))
    plasma.append(Species('proton',   n=1e11, T=1000))
    assert(debye(plasma[0])==plasma[0].debye)
    assert(debye(plasma) == approx(0.004879671013271479))

def test_negative_temperature(caplog):
    sp = Species(n=1e11, T=-1)
    assert(caplog.records[0].levelname == 'WARNING')



"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
from langmuir import *
from pytest import approx
import pytest
from scipy.constants import value as constants
try:
    from inspect import getfullargspec as getargspec # Python 3
except ImportError:
    from inspect import getargspec # Python 2

current_models = [
    thermal_current,
    normalization_current,
    OML_current,
    finite_radius_current,
    finite_length_current,
    finite_length_current_density
]

@pytest.fixture
def electron():
    return Species(n=1e11, T=1000)

@pytest.fixture
def proton():
    return Species('proton', n=1e11, T=1000)

@pytest.mark.parametrize("current", current_models)
def test_multiple_species(current, electron, proton):

    # Add a voltage for those models having a V argument.
    # This must be small to not make repelled current negligible.
    args = getargspec(current).args
    kwargs = {}
    if 'V' in args: kwargs['V'] = 0.1

    # geo = Sphere(electron.debye)
    geo = Cylinder(electron.debye, 10*electron.debye)
    Ie = current(geo, electron, **kwargs)
    Ii = current(geo, proton, **kwargs)
    I  = current(geo, [electron, proton], **kwargs)

    assert Ie+Ii == approx(I)

@pytest.mark.parametrize("current", current_models)
def test_normalization_thmax(current, electron):

    if current != normalization_current:

        args = getargspec(current).args
        kwargs = {}
        if 'V' in args: kwargs['V'] = 0.1

        geo = Cylinder(0.1*electron.debye, 1)
        In = current(geo, electron, normalization='thmax', **kwargs)
        I  = current(geo, electron, **kwargs)
        I0 = normalization_current(geo, electron)

        assert I0*In == approx(I)

@pytest.mark.parametrize("current", current_models)
def test_normalization_th(current, electron):

    if current not in [normalization_current, thermal_current]:

        args = getargspec(current).args
        kwargs = {}
        if 'V' in args: kwargs['V'] = 0.1

        geo = Cylinder(0.1*electron.debye, 1)
        In = current(geo, electron, normalization='th', **kwargs)
        I  = current(geo, electron, **kwargs)
        I0 = thermal_current(geo, electron)

        assert I0*In == approx(I)

@pytest.mark.parametrize("current", current_models)
def test_normalization_oml(current, electron):

    if current not in [normalization_current, thermal_current, OML_current]:

        args = getargspec(current).args
        kwargs = {}
        if 'V' in args: kwargs['V'] = 0.1

        geo = Cylinder(0.1*electron.debye, 1)
        In = current(geo, electron, normalization='oml', **kwargs)
        I  = current(geo, electron, **kwargs)
        I0 = OML_current(geo, electron, 0.1)

        assert I0*In == approx(I)

@pytest.mark.parametrize("current", current_models)
def test_eta(current, electron):

    args = getargspec(current).args
    kwargs = {}
    if 'V' in args or 'eta' in args:

        # Should have both or none
        assert 'V' in args
        assert 'eta' in args

        e = constants('elementary charge')
        k = constants('Boltzmann constant')
        T = 1000
        eta = -10

        geometry = Cylinder(r=electron.debye, l=10*electron.debye)
        I_eta = current(geometry, electron, eta=eta)
        I_V   = current(geometry, electron, V=-eta*k*T/e)

        assert I_eta == approx(I_V)

@pytest.mark.parametrize("current", current_models)
def test_multiple_species_eta(current, electron, proton, caplog):

    args = getargspec(current).args
    if 'eta' in args:

        I = current(Sphere(electron.debye), [electron, proton], eta=1)
        assert(caplog.records[0].levelname == 'ERROR')

@pytest.mark.parametrize("current", current_models)
def test_multiple_species_normalize(current, electron, proton, caplog):

    args = getargspec(current).args
    kwargs = {}
    geometry = Cylinder(r=electron.debye, l=10*electron.debye)
    if 'V' in args: kwargs['V'] = 0.1
    if 'normalization' in args:

        I = current(geometry, [electron, proton],
                    normalization='thmax', **kwargs)
        assert(caplog.records[0].levelname == 'ERROR')

@pytest.mark.parametrize("current", current_models)
def test_multiple_species_zeta(current, electron, proton, caplog):

    args = getargspec(current).args
    kwargs = {}
    geometry = Cylinder(r=electron.debye, l=10*electron.debye)
    if 'V' in args: kwargs['V'] = 0.1
    if 'zeta' in args:

        I = current(geometry, [electron, proton], zeta=39, **kwargs)
        assert(caplog.records[0].levelname == 'ERROR')

@pytest.mark.parametrize("current", current_models)
def test_input_output_format(current, electron):

    geo = Cylinder(r=0.1*electron.debye, l=10*electron.debye)
    # geo = Sphere(electron.debye)

    args = getargspec(current).args
    if 'V' in args:

        I = current(geo, electron, 0.1)
        assert(isinstance(I, float))

        I = current(geo, electron, 1)
        assert(isinstance(I, float))

        I = current(geo, electron, [1, 2])
        assert(isinstance(I, np.ndarray))
        assert(I.dtype==np.float)
        assert(I.shape==(2,))

        I = current(geo, electron, np.array([1, 2]))
        assert(isinstance(I, np.ndarray))
        assert(I.dtype==np.float)
        assert(I.shape==(2,))

@pytest.mark.parametrize("current", current_models)
def test_geometry_error(current, electron):

    args = getargspec(current).args
    kwargs = {}
    if 'V' in args: kwargs['V'] = 0.1

    # Many functions will fail when obtaining the
    # normalization current, which means to test the
    # current function's own fail guard you must not
    # use the normalization current.
    if 'normalization' in args: kwargs['normalization'] = 'thmax'

    with pytest.raises(ValueError):
        current('Bullshit', electron, **kwargs)

@pytest.mark.parametrize("current", current_models)
def test_normalization_error(current, electron):

    args = getargspec(current).args
    kwargs = {}
    if 'eta' in args: kwargs['eta'] = -1
    if 'normalization' in args:
        kwargs['normalization'] = 'Bullshit'

        geo = Cylinder(electron.debye, electron.debye)
        with pytest.raises(ValueError):
            current(geo, electron, **kwargs)



"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
from langmuir import *
from pytest import approx
import pytest

# I generally try to test one point in each table, with unequal row/column
# indices and where nearby elements differ from the one I test.

@pytest.fixture
def electron():
    return Species(n=1e11, T=1000)

def test_laframboise_sphere(electron):
    I = finite_radius_current(Sphere(electron.debye), electron, eta=-0.6,
                              normalization='thmax', table='laframboise')
    assert(I == approx(1.595))

def test_laframboise_cylinder(electron):
    I = finite_radius_current(Cylinder(3*electron.debye, 1), electron, eta=-2.0,
                              normalization='thmax', table='laframboise')
    assert(I == approx(1.928))

def test_darian_marholm_uncomplete_sphere(electron):

    def curr(kappa, alpha):
        sp = Species(n=1e11, T=1000, kappa=kappa, alpha=alpha)
        return finite_radius_current(Sphere(2*sp.debye), sp, eta=-1,
                                     normalization='thmax',
                                     table='darian-marholm uncomplete')

    # Testing one element for each alpha/kappa table
    assert(curr(float('inf'), 0  ) == approx(1.958))
    assert(curr(float('inf'), 0.2) == approx(2.069))
    assert(curr(6           , 0  ) == approx(1.982))
    assert(curr(6           , 0.2) == approx(2.369))
    assert(curr(4           , 0  ) == approx(1.990))
    assert(curr(4           , 0.2) == approx(2.846))

    # Testing that V=0 is the inaccurate one and that R=0 doesn't exist
    I = finite_radius_current(Sphere(0.2*electron.debye), electron, 0,
                              normalization='thmax', table='darian-marholm uncomplete')
    assert(I == approx(0.965))

def test_darian_marholm_sphere(electron):

    # Testing one from the uncomplete subset
    sp = Species(n=1e11, T=1000, kappa=6)
    I = finite_radius_current(Sphere(2*sp.debye), sp, eta=-1,
                              normalization='thmax', table='darian-marholm')
    assert(I == approx(1.982))

    # Testing that V=0 is the accurate one (not the one from uncomplete)
    I = finite_radius_current(Sphere(0.2*electron.debye), electron, eta=0,
                              normalization='thmax')
    assert(I == approx(1.0))

    # Testing that R=0 exists
    I = finite_radius_current(Sphere(0), electron, eta=-3,
                              normalization='thmax', table='darian-marholm')
    assert(I == approx(4.0))

def test_darian_marholm_uncomplete_cylinder(electron):

    def curr(kappa, alpha):
        sp = Species(n=1e11, T=1000, kappa=kappa, alpha=alpha)
        return finite_radius_current(Cylinder(3*sp.debye, 1), sp, eta=-1,
                                     normalization='thmax',
                                     table='darian-marholm uncomplete')

    assert(curr(float('inf'), 0  ) == approx(1.538))
    assert(curr(float('inf'), 0.2) == approx(1.914))
    assert(curr(6           , 0  ) == approx(1.509))
    assert(curr(6           , 0.2) == approx(2.165))
    assert(curr(4           , 0  ) == approx(1.510))
    assert(curr(4           , 0.2) == approx(2.547))

    # Testing that V=0 is the inaccurate one and that R=0 doesn't exist
    I = finite_radius_current(Cylinder(1.0*electron.debye, 1), electron, 0,
                              normalization='thmax', table='darian-marholm uncomplete')
    assert(I == approx(0.974))

def test_darian_marholm_cylinder(electron):

    # Testing one from the uncomplete subset
    sp = Species(n=1e11, T=1000, kappa=6)
    I = finite_radius_current(Cylinder(3*sp.debye, 1), sp, eta=-1,
                              normalization='thmax', table='darian-marholm')
    assert(I == approx(1.509))

    # Testing that V=0 is the accurate one (not the one from uncomplete)
    I = finite_radius_current(Cylinder(electron.debye, 1), electron, eta=0,
                              normalization='thmax')
    assert(I == approx(1.0))

    # Testing that R=0 exists
    I = finite_radius_current(Cylinder(0,1), electron, eta=-3,
                              normalization='thmax', table='darian-marholm')
    assert(I == approx(2.2417, 1e-3))

def test_laframboise_darian_marholm_sphere(electron):

    # Testing one from each of Darian-Marholm and Laframboise

    sp = Species(n=1e11, T=1000, kappa=4, alpha=0.2)
    I = finite_radius_current(Sphere(3*sp.debye), sp, eta=-5, normalization='thmax',
                              table='laframboise-darian-marholm')
    assert(I == approx(4.535))

    I = finite_radius_current(Sphere(3*electron.debye), electron, eta=-5,
                              normalization='thmax',
                              table='laframboise-darian-marholm')
    assert(I == approx(4.640))

def test_laframboise_darian_marholm_cylinder(electron):

    # Testing one from each of Darian-Marholm and Laframboise
    sp = Species(n=1e11, T=1000, kappa=4, alpha=0.2)
    I = finite_radius_current(Cylinder(3*sp.debye, 1), sp, eta=-5,
                              normalization='thmax',
                              table='laframboise-darian-marholm')
    assert(I == approx(3.465))

    I = finite_radius_current(Cylinder(3*electron.debye, 1), electron, eta=-5,
                              normalization='thmax',
                              table='laframboise-darian-marholm')
    assert(I == approx(2.701))

def test_finite_radius_vs_OML():

    sp = Species(n=1e11, T=1000, alpha=0.2, kappa=6)
    geo = Sphere(0.2*sp.debye)
    I_OML = OML_current(geo, sp, eta=-15)
    I_fr = finite_radius_current(geo, sp, eta=-15)
    assert(I_OML == approx(I_fr, 0.03))

def test_discard_spectral_indices(caplog):

    sp = Species(n=1e11, T=1000, alpha=0.2, kappa=6)
    geo = Sphere(0.2*sp.debye)

    I = finite_radius_current(geo, sp, eta=-15, table='laframboise')
    assert(len(caplog.records) == 1)

def test_discard_repelled_particles(caplog):

    electron = Species(n=1e11, T=1000)
    I = finite_radius_current(Sphere(electron.debye), electron, eta=[-1,1])
    assert(len(caplog.records) == 1)
    assert('negative' in caplog.text)
    assert('positive' not in caplog.text)

    ion = Species('proton', n=1e11, T=1000)
    I = finite_radius_current(Sphere(ion.debye), ion, eta=[-1,1])
    assert(len(caplog.records) == 2)
    assert('positive' in caplog.text)

def test_domain(caplog, electron):

    # Too high voltage
    I = finite_radius_current(Sphere(electron.debye), electron, eta=-30)
    assert(np.isnan(I))

    # Too large radius
    I = finite_radius_current(Sphere(200*electron.debye), electron, eta=-20)
    assert(np.isnan(I))

    # Too high alpha
    sp = Species(n=1e11, T=1000, alpha=0.3)
    I = finite_radius_current(Sphere(sp.debye), sp, eta=-20)
    assert(np.isnan(I))

    # Too low kappa
    sp = Species(n=1e11, T=1000, kappa=0.2)
    I = finite_radius_current(Sphere(sp.debye), sp, eta=-20)
    assert(np.isnan(I))

    assert(len(caplog.records) == 4)



"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
from langmuir.tables import *
from langmuir.geometry import *
from langmuir.species import *
from langmuir.misc import *
from scipy.interpolate import griddata
from scipy.constants import value as constants
from copy import deepcopy
import scipy.special as special
import numpy as np
import os

def finite_radius_current(geometry, species, V=None, eta=None, normalization=None,
                          table='laframboise-darian-marholm'):
    """
    A current model taking into account the effects of finite radius by
    interpolating between tabulated normalized currents. The model only
    accounts for the attracted-species currents (for which eta<0). It does
    not extrapolate, but returns ``nan`` when the input parameters are outside
    the domain of the model. This happens when the normalized potential for any
    given species is less than -25, when kappa is less than 4, when alpha is
    more than 0.2 or when the radius is more than 10 or sometimes all the way
    up towards 100 (as the distribution approaches Maxwellian). Normally finite
    radius effects are negligible for radii less than 0.2 Debye lengths (spheres)
    or 1.0 Debye lengths (cylinders).

    The model can be based on the following tables, as decided by the ``table``
    parameter:

    - ``'laframboise'``.
      The de-facto standard tables for finite radius currents, tables 5c
      and 6c in Laframboise, "Theory of Spherical and Cylindrical Langmuir
      Probes in a Collisionless, Maxwellian Plasma at Rest", PhD Thesis.
      Covers Maxwellian plasmas only, probe radii ranging from 0 to 100 Debye
      lengths.

    - ``'darian-marholm uncomplete'``.
      These tables covers Maxwellian, Kappa, Cairns and Kappa-Cairns
      distributions for radii ranging from 0.2 Debye lengths (spheres) or
      1.0 Debye length (cylinders) up to 10 Debye lengths. They are not as
      accurate as ``'laframboise'`` for pure the Maxwellian, but usually within
      one or two percent.

    - ``'darian-marholm'``.
      Same as above, but this is complemented by adding analytical values from
      OML theory, thereby extending the range of valid radii down to zero Debye
      lengths. In addition, the values for zero potential are replaced by
      analytical values (i.e. the thermal current), since these are amongst the
      most inaccurate in the above, and more accurate values can be analytically
      computed.

    - ``'laframboise-darian-marholm'``.
      This replaces the tabulated values for the Maxwellian distribution in
      ``'darian-marholm'`` with those of Laframboise. Accordingly this table
      produces the most accurate result available in any situation, and has the
      widest available parameter domain, with the probe radius gradually
      increasing from 10 towards 100 Debye lengths as the distribution
      approaches the Maxwellian.

    Parameters
    ----------
    geometry: Plane, Cylinder or Sphere
        Probe geometry

    species: Species or array-like of Species
        Species constituting the background plasma

    V: float or array-like of floats
        Probe voltage(s) in [V]. Overrides eta.

    eta: float or array-like of floats
        Probe voltage(s) normalized by k*T/q, where q and T are the species'
        charge and temperature and k is Boltzmann's constant.

    normalization: 'th', 'thmax', 'oml', None
        Wether to normalize the output current by, respectively, the thermal
        current, the Maxwellian thermal current, the OML current, or not at
        all, i.e., current in [A/m].

    table: string
        Which table to use for interpolation. See detailed description above.

    Returns
    -------
    float if voltage is float. array of floats corresponding to voltage if
    voltage is array-like.
    """
    if isinstance(species, list):
        if normalization is not None:
            logger.error('Cannot normalize current to more than one species')
            return None
        if eta is not None:
            logger.error('Cannot normalize voltage to more than one species')
            return None
        I = 0
        for s in species:
            I += finite_radius_current(geometry, s, V, eta, table=table)
        return I

    q, m, n, T = species.q, species.m, species.n, species.T
    kappa, alpha = species.kappa, species.alpha
    k = constants('Boltzmann constant')

    if V is not None:
        V = make_array(V)
        eta = q*V/(k*T)
    else:
        eta = make_array(eta)

    eta = deepcopy(eta)

    I = np.zeros_like(eta)

    indices_n = np.where(eta > 0)[0]   # indices for repelled particles
    indices_p = np.where(eta <= 0)[0]  # indices for attracted particles

    if normalization is None:
        I0 = normalization_current(geometry, species)
    elif normalization.lower() == 'thmax':
        I0 = 1
    elif normalization.lower() == 'th':
        I0 = normalization_current(geometry, species)/\
             thermal_current(geometry, species)
    elif normalization.lower() == 'oml':
        I0 = normalization_current(geometry, species)/\
             OML_current(geometry, species, eta=eta)
    else:
        raise ValueError('Normalization not supported: {}'.format(normalization))

    if isinstance(geometry, Sphere):
        table += ' sphere'
    elif isinstance(geometry, Cylinder):
        table += ' cylinder'
    else:
        raise ValueError('Geometry not supported: {}'.format(geometry))

    R = geometry.r/species.debye

    if "darian-marholm" in table:
        table = get_table(table)
        pts = table['points']
        vals = table['values'].reshape(-1)
        I[indices_p] = I0*griddata(pts, vals, (1/kappa, alpha, R, eta[indices_p]))

    else:
        table = get_table(table)
        pts = table['points']
        vals = table['values'].reshape(-1)
        I[indices_p] = I0*griddata(pts, vals, (R, eta[indices_p]))
        if(kappa != float('inf') or alpha != 0):
            logger.warning("Using pure Laframboise tables discards spectral indices kappa and alpha")

    if len(indices_n)>0:
        pos_neg = "positive" if q>0 else "negative"
        logger.warning("Only attracted species current is covered by tabulated "
                       "values. Currents due to {} is set to zero for "
                       "{} potentials".format(species, pos_neg))

    if any(np.isnan(I)):
        logger.warning("Data points occurred outside the domain of tabulated values resulting in nan")

    return I[0] if len(I) == 1 else I



"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""

from scipy.constants import value as constants
from langmuir.misc import *
import numpy as np

class Species(object):
    """
    Defines a species using a set of flags and keyword parameters. Maxwellian
    electrons are the default but Kappa, Cairns, and Kappa-Cairns distributions
    can be specified by setting the spectral indices kappa and alpha. Other
    parameters can be specified/overridden using the keyword parameters.
    Examples::

        >>> # Maxwellian electrons with density 1e11 m^(-3) and 1000 K
        >>> species = Species(n=1e11, T=1000)

        >>> # Kappa-distributed electrons with kappa=2
        >>> species = Species(n=1e11, T=1000, kappa=2)

        >>> # Doubly charged Maxwellain Oxygen ions with 0.26 eV
        >>> species = Species(Z=2, amu=16, n=1e11, eV=0.26)

        >>> # Kappa-Cairns-distributed protons with kappa = 5 and alpha=0.2
        >>> species = Species('proton', n=1e11, T=1000, kappa=5, alpha=0.2)

    A plasma is fully specified by a list of species. E.g. for a Maxwellian
    electron-proton plasma::

        >>> plasma = []
        >>> plasma.append(Species('electron', n=1e11, T=1000))
        >>> plasma.append(Species('proton',   n=1e11, T=1000))

    Flags:
    ------
    'electron'     : Elecron species (default)
    'proton'       : Proton species
    'positron'     : Positron species

    Keyword parameters:
    -------------------
    m     : mass [kg]
    amu   : mass [amu]
    q     : charge [C]
    Z     : charge [elementary charges]
    n     : density [m^(-3)]
    T     : temperature [K]
    eV    : temperature [eV]
    vth   : thermal velocity [m/s]
    kappa : spectral index kappa for Kappa and Kappa-Cairns distribution
    alpha : spectral index alpha for Cairns and Kappa-Cairns distribution
    """

    def __init__(self, *args, **kwargs):

        e    = constants('elementary charge')
        me   = constants('electron mass')
        mp   = constants('proton mass')
        k    = constants('Boltzmann constant')
        eps0 = constants('electric constant')
        amu  = constants('atomic mass constant')

        if 'Z' in kwargs:
            self.q = kwargs['Z']*e
        else:
            self.q = kwargs.pop('q', -e)

        if 'amu' in kwargs:
            self.m = kwargs['amu']*amu
        else:
            self.m = kwargs.pop('m', me)

        for arg in args:
            if arg.lower()=='proton':
                self.q = e
                self.m = mp

            if arg.lower()=='positron':
                self.q = e
                self.m = me

        self.Z = self.q/e
        self.amu = self.m/amu

        self.n = kwargs['n']

        if 'eV' in kwargs:
            self.T = kwargs['eV']*e/k
        elif 'vth' in kwargs:
            self.T = self.m*kwargs['vth']**2/k
        else:
            self.T = kwargs['T']

        if self.T<0:
            self.T=0
            logger.warning('Negative temperature interpreted as zero')


        self.vth = np.sqrt(k*self.T/self.m)
        self.eV  = self.T*k/e
        self.omega_p = np.sqrt(self.q**2*self.n/(eps0*self.m))
        self.freq_p = self.omega_p/(2*np.pi)
        self.period_p = 1/self.freq_p

        self.alpha = kwargs.pop('alpha', 0)
        self.kappa = kwargs.pop('kappa', float('inf'))

        if self.kappa == float('inf'):
            self.debye = np.sqrt(eps0*k*self.T/(self.q**2*self.n))*np.sqrt((1.0 + 15.0*self.alpha)/(1.0 + 3.0*self.alpha))
        else:
            self.debye = np.sqrt(eps0 * k * self.T / (self.q**2 * self.n)) *\
                         np.sqrt(((self.kappa - 1.5) / (self.kappa - 0.5)) *\
                        ((1.0 + 15.0 * self.alpha * ((self.kappa - 1.5) / (self.kappa - 2.5))) / (1.0 + 3.0 * self.alpha * ((self.kappa - 1.5) / (self.kappa - 0.5)))))

    def omega_c(self, B):
        """
        Returns the angular cyclotron frequency of the species for a given B-field
        """
        return np.abs(B*self.q/self.m)

    def freq_c(self, B):
        """
        Returns the cyclotron frequency of the species for a given B-field
        """
        return self.omega_c(B)/(2*np.pi)

    def period_c(self, B):
        """
        Returns the cyclotron period of the species for a given B-field
        """
        return 1/self.freq_c(B)

    def larmor(self, B):
        """
        Returns the Larmor radius of the species for a given B-field
        """
        return self.vth/self.omega_c(B)

    def __repr__(self):

        s = "Species(q={}, m={}, n={}, T={}".format(self.q, self.m, self.n, self.T)

        if(self.kappa != float('inf')):
            s += ", kappa={}".format(self.kappa)

        if(self.alpha != 0):
            s += ", alpha={}".format(self.alpha)

        s += ")"

        return s

def debye(species):
    """
    Get the total Debye length of a species or a plasma (represented as a list
    of species). If species is a list this includes the Debye length due to
    all species.
    """
    if not isinstance(species, list):
        species = [species]
    return sum([s.debye**(-2) for s in species])**(-0.5)



"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
import logging
import numpy as np

logger = logging.getLogger('langmuir')
logging.basicConfig()

def make_array(arr):
    """
    Takes a list, tuple, integer or float and returns a numpy array.
    """
    if isinstance(arr, (int, float, np.int64, np.int32, np.float64, np.float32)):
        arr = [arr]
    return np.array(arr, dtype=np.float)




"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
from langmuir import *
from pytest import approx
import pytest

@pytest.fixture
def electron():
    return Species(n=1e11, T=1000)

@pytest.fixture
def eta_fl():
    return np.array([2, 10, 17, 25, 32, 50, 75, 100], dtype=np.float)

def test_current_normalize_OML(electron, eta_fl):

    geo = Cylinder(r=0.1*electron.debye, l=10*electron.debye)
    g = finite_length_current(geo, electron, eta=-eta_fl, normalization='OML')
    I = finite_length_current(geo, electron, eta=-eta_fl)
    I0 = OML_current(geo, electron, eta=-eta_fl)

    assert np.allclose(g, I/I0)

def test_current_normalize_th(electron, eta_fl):

    geo = Cylinder(r=0.1*electron.debye, l=10*electron.debye)
    g = finite_length_current(geo, electron, eta=-eta_fl, normalization='th')
    I = finite_length_current(geo, electron, eta=-eta_fl)
    I0 = thermal_current(geo, electron)

    assert np.allclose(g, I/I0)

def test_current_density_normalize_OML(electron, eta_fl):

    geo = Cylinder(r=0.1*electron.debye, l=10*electron.debye)
    geonorm = Cylinder(r=0.1*electron.debye, l=1)
    zeta = np.linspace(1,10,10)
    for eta in eta_fl:
        g = finite_length_current_density(geo, electron, eta=-eta, zeta=zeta, normalization='OML')
        i = finite_length_current_density(geo, electron, eta=-eta, zeta=zeta)
        i0 = OML_current(geonorm, electron, eta=-eta)
        gf = i/i0

        assert np.allclose(g, gf)

def test_current_density_normalize_th(electron, eta_fl):

    geo = Cylinder(r=0.1*electron.debye, l=10*electron.debye)
    geonorm = Cylinder(r=0.1*electron.debye, l=1)
    zeta = np.linspace(1,10,10)
    for eta in eta_fl:
        g = finite_length_current_density(geo, electron, eta=-eta, zeta=zeta, normalization='th')
        i = finite_length_current_density(geo, electron, eta=-eta, zeta=zeta)
        i0 = thermal_current(geonorm, electron)
        gf = i/i0

        assert np.allclose(g, gf)

def test_current_vs_OML(electron):

    # Agrees to within 5% of OML for infinite probes

    geo = Cylinder(r=electron.debye, l=10*electron.debye,
                  # lguard=float('inf'), rguard=500*electron.debye)
                  lguard=float('inf'), rguard=float('inf'))
    for eta in [2, 6, 10, 17, 25, 32, 50, 75, 100]:
        I = finite_length_current(geo, electron, eta=-eta, normalization='OML')
        assert np.allclose(I, 1, 0.05, 0), "eta={}".format(-eta)

def test_current_density_vs_OML(electron):

    # Agrees to within 5% of OML for infinite probes

    geo = Cylinder(r=electron.debye, l=1,
                   lguard=float('inf'), rguard=float('inf'))
    for eta in [2, 6, 10, 17, 25, 32, 50, 75, 100]:
        I_OML = OML_current(geo, electron, eta=-eta)
        I_fl = finite_length_current_density(geo, electron, eta=-eta, z=0.5)
        assert I_OML == approx(I_fl, 0.05), "eta={}".format(-eta)

def test_current_guard(electron):

    r = electron.debye
    l = 10*electron.debye
    long = 1000*electron.debye
    eta = 17

    Ino   = finite_length_current(Cylinder(r, l), electron, eta=-eta)

    Iinf  = finite_length_current(Cylinder(r, l, lguard=np.inf), electron, eta=-eta)
    Ilong = finite_length_current(Cylinder(r, l, lguard=long), electron, eta=-eta)
    assert Iinf == approx(Ilong)
    assert not Iinf == approx(Ino)

    Iinf  = finite_length_current(Cylinder(r, l, rguard=np.inf), electron, eta=-eta)
    Ilong = finite_length_current(Cylinder(r, l, rguard=long), electron, eta=-eta)
    assert Iinf == approx(Ilong)
    assert not Iinf == approx(Ino)

    Iinf  = finite_length_current(Cylinder(r, l, lguard=l, rguard=np.inf), electron, eta=-eta)
    Ilong = finite_length_current(Cylinder(r, l, lguard=l, rguard=long), electron, eta=-eta)
    assert Iinf == approx(Ilong)

def test_current_density_guard(electron):

    r = electron.debye
    l = 10*electron.debye
    long = 1000*electron.debye
    eta = 17

    Ino   = finite_length_current_density(Cylinder(r, l), electron, eta=-eta)

    Iinf  = finite_length_current_density(Cylinder(r, l, lguard=np.inf), electron, eta=-eta)
    Ilong = finite_length_current_density(Cylinder(r, l, lguard=long), electron, eta=-eta)
    assert Iinf == approx(Ilong)
    assert not Iinf == approx(Ino)

    Iinf  = finite_length_current_density(Cylinder(r, l, rguard=np.inf), electron, eta=-eta)
    Ilong = finite_length_current_density(Cylinder(r, l, rguard=long), electron, eta=-eta)
    assert Iinf == approx(Ilong)
    assert not Iinf == approx(Ino)

    Iinf  = finite_length_current_density(Cylinder(r, l, lguard=l, rguard=np.inf), electron, eta=-eta)
    Ilong = finite_length_current_density(Cylinder(r, l, lguard=l, rguard=long), electron, eta=-eta)
    assert Iinf == approx(Ilong)

def test_current_samples():

    # Testing that a few random sample points don't change. All points are in
    # the original data, so the test should be independent of interpolation

    elec = Species(n=35e10, eV=0.08)

    r = 1e-3
    ls = np.array([5, 30, 80, 400, 2000], dtype=np.float)*1e-3
    etas = np.array([100, 2, 32, 17, 25], dtype=np.float)
    Ifs = np.array([-7.311843622376143e-06,
                    -1.2270514750624203e-06,
                    -1.332414356159239e-05,
                    -3.5970798600771154e-05,
                    -0.00020182061197445252])

    for l, eta, If in zip(ls, etas, Ifs):
        geo = Cylinder(r=r, l=l)
        I = finite_length_current(geo, elec, eta=-eta)
        assert I == approx(If)

def test_current_density_samples():

    # Testing that a few random sample points don't change. All points are in
    # the original data, so the test should be independent of interpolation

    elec = Species(n=35e10, eV=0.08)

    r = 1e-3
    ls = np.array([5, 30, 80, 400, 2000], dtype=np.float)*1e-3
    etas = np.array([100, 2, 32, 17, 25], dtype=np.float)
    zs = np.array([0, .22, .1, .75, 0.91, 1])
    Ifs = np.array([[-0.00126263, -0.00147027, -0.00137461, -0.00148828, -0.00136481, -0.00126263],
                    [-4.02337566e-05, -4.10898286e-05, -4.26822956e-05, -4.07257345e-05, -4.27430040e-05, -4.02337566e-05],
                    [-0.00021501, -0.00013982, -0.00021829, -0.00013476, -0.00023155, -0.00021501],
                    [-1.41492521e-04, -8.50936721e-05, -8.51884293e-05, -8.50936707e-05, -8.53236120e-05, -1.41492521e-04],
                    [-1.76726210e-04, -9.93476796e-05, -9.93476796e-05, -9.93476796e-05, -9.93476796e-05, -1.76726210e-04]])

    for l, eta, If in zip(ls, etas, Ifs):
        geo = Cylinder(r=r, l=l)
        I = finite_length_current_density(geo, elec, eta=-eta, z=zs*l)
        assert np.allclose(I, If)

def test_current_interpolated_samples():

    # Testing that a few random sample points don't change. These points depen
    # upon the interpolation scheme. 

    elec = Species(n=35e10, eV=0.08)

    r = 1e-3
    ls = np.array([7, 70, 700], dtype=np.float)*1e-3
    etas = np.array([65, 1, 90], dtype=np.float)

    # Ground truth when using scipy.interpolate.griddata
    # Ifs = np.array([-6.2794005496528275e-06,
    #                 -1.9760221828553347e-06,
    #                 -0.00014456804182897328])

    # Ground truth when using 1st order scipy.interpolate.RectBivariateSpline
    Ifs = np.array([-6.439906296987808e-06,
                    -1.9562053783242e-06,
                    -0.00014420124702985558])

    for l, eta, If in zip(ls, etas, Ifs):
        geo = Cylinder(r=r, l=l)
        I = finite_length_current(geo, elec, eta=-eta, interpolate='g')
        assert I == approx(If)

    # Ground truth with linear interpolation in I
    Ifs = np.array([-5.951577246367211e-06,
                    -1.9530626386338544e-06,
                    -0.00014358695303707134])

    for l, eta, If in zip(ls, etas, Ifs):
        geo = Cylinder(r=r, l=l)
        I = finite_length_current(geo, elec, eta=-eta, interpolate='I')
        assert I == approx(If)

def test_current_density_interpolated_samples():

    # Testing that a few random sample points don't change. These points depen
    # upon the interpolation scheme. 

    elec = Species(n=35e10, eV=0.08)

    r = 1e-3
    ls = np.array([7, 70, 700], dtype=np.float)*1e-3
    etas = np.array([65, 1, 90], dtype=np.float)
    zs = np.array([0, .4, .29, 1])

    # Ground truth when using scipy.interpolate.griddata
    # Ifs = np.array([[-0.00045615, -0.00106614, -0.00099727, -0.00045615],
    #                 [-2.92817683e-05, -2.75155255e-05, -2.76705163e-05, -2.92817683e-05],
    #                 [-0.00036732, -0.00019135, -0.00019135, -0.00036732]])

    # Ground truth when using 1st order scipy.interpolate.RectBivariateSpline
    Ifs = np.array([[-0.00054336, -0.00105909, -0.00100473, -0.00054336],
                    [-2.89516395e-05, -2.72642938e-05, -2.74123613e-05, -2.89516395e-05],
                    [-0.00036957, -0.00019082, -0.00019082, -0.00036957]])

    for l, eta, If in zip(ls, etas, Ifs):
        geo = Cylinder(r=r, l=l)
        I = finite_length_current_density(geo, elec, eta=-eta, z=zs*l)
        assert np.allclose(I, If)

def test_current_non_maxwellian(caplog):

    elec = Species(n=35e10, eV=0.08)
    geo = Cylinder(0.1*elec.debye, 10*elec.debye)
    I = finite_length_current(geo, Species(n=35e10, T=1000, alpha=0.3), 3.0)
    I = finite_length_current(geo, Species(n=35e10, T=1000, kappa=5), 3.0)

    assert(len(caplog.records) == 2)

def test_current_invalid_interpolate(caplog):

    elec = Species(n=35e10, eV=0.08)
    geo = Cylinder(0.1*elec.debye, 10*elec.debye)

    with pytest.raises(ValueError):
        finite_length_current(geo, Species(n=35e10, T=1000), 3.0, interpolate='a')

def test_current_too_high_eta(caplog):

    elec = Species(n=35e10, eV=0.08)
    geo = Cylinder(0.1*elec.debye, 10*elec.debye)
    I = finite_length_current(geo, elec, eta=-111)
    assert(len(caplog.records) == 1)

def test_current_density_non_maxwellian(caplog):

    elec = Species(n=35e10, eV=0.08)
    geo = Cylinder(0.1*elec.debye, 10*elec.debye)
    I = finite_length_current_density(geo, Species(n=35e10, T=1000, alpha=0.3), 3.0)
    I = finite_length_current_density(geo, Species(n=35e10, T=1000, kappa=5), 3.0)

    assert(len(caplog.records) == 2)

def test_current_density_matrix(electron):

    geo = Cylinder(0.1*electron.debye, 10*electron.debye)
    eta = np.array([-1, 0, 1, 10])
    zeta = np.linspace(0, 10, 8)
    I_matrix = finite_length_current_density(geo, electron, eta=-eta, zeta=zeta)
    for i, e in enumerate(eta):
        I_row = finite_length_current_density(geo, electron, eta=-e, zeta=zeta)
        assert np.allclose(I_row, I_matrix[i])

def test_current_repelled(electron):

    geo = Cylinder(0.1*electron.debye, electron.debye)
    V = -np.array([0.1, 1.2, 56])
    I = finite_length_current(geo, electron, V)
    I_OML = OML_current(geo, electron, V)
    assert np.allclose(I, I_OML)

def test_current_density_repelled(electron):

    geo = Cylinder(0.1*electron.debye, 1)
    V = -np.array([0.1, 1.2, 56])
    I = finite_length_current_density(geo, electron, V)
    I_OML = OML_current(geo, electron, V)
    assert np.allclose(I, I_OML)
